from configparser import ConfigParser


config_file: str = ".config"

config: ConfigParser = ConfigParser()
config.read(filenames=config_file)
configs: dict = {}
try:
    configs = dict(config.items(section="DATABASE"))
except:
    pass

DB_HOST = configs.get("db_host")
DB_PORT = configs.get("db_port") 
DB_NAME = configs.get("db_name")
DB_USER = configs.get("db_user")
DB_PASSWORD = configs.get("db_password")

DNS = "host={} port={} dbname={} user={} password={}".format(DB_HOST, DB_PORT, DB_NAME, DB_USER, DB_PASSWORD)



"""
Lightweight and simple ORM for Python (with psycopg)
"""
__version__ = "1.0.0"
from .config import DNS
from .base import DataBase
from .types import (
    BigInt,
    BigSerial,
    Boolean,
    Date,
    Float,
    Integer,
    JSON,
    Password,
    Real,
    Serial,
    SmallInt,
    SmallSerial,
    Text,
    Time,
    TimeStamp,
    UUID,
    VarChar,
    XML,
    Model,
)

__all__ = [
    "DNS",
    "DataBase",
    "Model",
    "BigInt",
    "BigSerial",
    "Boolean",
    "Date",
    "Float",
    "Integer",
    "JSON",
    "Password",
    "Real",
    "Serial",
    "SmallInt",
    "SmallSerial",
    "Text",
    "Time",
    "TimeStamp",
    "UUID",
    "VarChar",
    "XML",
]



import json
import re
from psycopg import sql
from psycopg.rows import class_row
from dataclasses import dataclass
from psycopg import errors

from kirpi.base import DataBase
from kirpi.utils import parse_filters, hash_password


class OrderBy:
    def __init__(self, *columns) -> None:
        self.columns = columns or ["id__asc"]

    def generate(self):
        order_by_columns = {}
        for i in self.columns:
            if len(i.split("__")) == 2:
                order_by_columns[i.split("__")[0]] = i.split("__")[1].upper()
        return order_by_columns

class Model:
    def __init__(self):
        self.name = self.__class__.__name__.lower() + "s"
    
    def parse_fields(self):
        fields = []
        dict_data: dict = self.__class__.__dict__
        for field in dict_data:
            if isinstance(dict_data[field], Type):
                fields.append("{} {}".format(field, dict_data[field].generate()))
        return fields
        
    def create_table(self):
        cursor = DataBase().context.cursor()
        query: sql.SQL = sql.SQL("CREATE TABLE IF NOT EXISTS {table_name} ({fields});").format(
            table_name=sql.Identifier(self.name),
            fields=sql.SQL(", ".join(self.parse_fields()))
        )
        # print(query.as_string(DataBase.context))
        cursor.execute(query=query)
        DataBase().context.commit()

        for col in self.fields():
            query: sql.SQL = sql.SQL(
                "DO $$"
                    " BEGIN"
                        " BEGIN"
                            " ALTER TABLE {} ADD COLUMN {} {};"
                        " EXCEPTION"
                            " WHEN duplicate_column THEN RAISE NOTICE 'column already exists';"
                        " END;"
                    " END;"
                " $$"
            ).format(
                sql.Identifier(self.name),
                sql.Identifier(col),
                sql.SQL(self.fields()[col].generate())
            )
            cursor.execute(query=query)
            DataBase().context.commit()
    
    def fields(self):
        fields = []
        dict_data: dict = self.__class__.__dict__
        for field in dict_data:
            if isinstance(dict_data[field], Type):
                fields.append(field)
        return fields
    
    @classmethod
    def execute(cls, query, schema: dataclass = None):
        cursor = None
        if schema:
            cursor = DataBase().context.cursor(row_factory=schema)
        else:
            cursor = DataBase().context.cursor()
        cursor.execute(query=query)
        DataBase().context.commit()
        return cursor.fet
    
    @classmethod
    def fields(cls):
        fields = {}
        dict_data: dict = cls.__dict__
        for field in dict_data:
            if isinstance(dict_data[field], Type):
                fields[field] = dict_data[field]
        return fields
    
    @classmethod
    def all(cls, page: int = 1, limit: int = 20, order_by: OrderBy = None, schema: dataclass = None):
        if order_by:
            if isinstance(order_by, OrderBy):
                pass
            else:
                order_by = OrderBy()
        else:
            order_by = OrderBy()
        page = page - 1
        fields: list = []
        if schema:
            for i in schema.__annotations__:
                if i == "id":
                    continue
                else:
                    fields.append(sql.Identifier(i))
        else:
            for i in cls.fields():
                if i == "id":
                    continue
                else:
                    fields.append(sql.Identifier(i))
        cursor = None
        if schema:
            cursor = DataBase().context.cursor(row_factory=class_row(schema))
        else:
            cursor = DataBase().context.cursor()
        query = sql.SQL("SELECT * FROM (SELECT ROW_NUMBER() OVER (ORDER BY id) AS \"id\", {fields} FROM {table}) x WHERE \"id\" > {page} ORDER BY {order_by} LIMIT {limit};").format(
            fields=sql.SQL(",").join(fields),
            table=sql.Identifier(cls.__name__.lower() + "s"),
            page=page,
            order_by=sql.SQL(", ".join([" ".join([i, order_by.generate()[i]]) for i in order_by.generate()])),
            limit=limit
        )
        cursor.execute(query=query)
        return cursor.fetchall()

    @classmethod
    def filter(cls, page: int = 1, limit: int = 20, order_by: OrderBy = None, schema: dataclass = None, **filters):
        if order_by:
            if isinstance(order_by, OrderBy):
                pass
            else:
                order_by = OrderBy()
        else:
            order_by = OrderBy()
        page = page - 1
        query = ""
        cursor = None
        if schema:
            cursor = DataBase().context.cursor(row_factory=class_row(schema))
        else:
            cursor = DataBase().context.cursor()
        fields: list = []
        if schema:
            for i in schema.__annotations__:
                if i == "id":
                    continue
                else:
                    fields.append(sql.Identifier(i))
        else:
            for i in cls.fields():
                if i == "id":
                    continue
                else:
                    fields.append(sql.Identifier(i))
        if parse_filters(**filters):
            query = sql.SQL("SELECT * FROM (SELECT ROW_NUMBER() OVER (ORDER BY id) AS \"id\", {fields} FROM {table}) x WHERE id > {page} AND {filters} ORDER BY {order_by} LIMIT {limit};").format(
                fields=sql.SQL(",").join(fields),
                table=sql.Identifier(cls.__name__.lower() + "s"),
                page=page,
                filters=sql.SQL(" AND ".join(list(map(lambda x: x.as_string(DataBase().context), parse_filters(**filters))))),
                order_by=sql.SQL(", ".join([" ".join([i, order_by.generate()[i]]) for i in order_by.generate()])),
                limit=limit
            )
        else:
            query = sql.SQL("SELECT * FROM (SELECT ROW_NUMBER() OVER (ORDER BY id) AS \"id\", {fields} FROM {table}) x WHERE id > {page} {filters} ORDER BY {order_by} LIMIT {limit};").format(
                fields=sql.SQL(",").join(fields),
                table=sql.Identifier(cls.__name__.lower() + "s"),
                page=page,
                filters=sql.SQL(" AND ".join(list(map(lambda x: x.as_string(DataBase().context), parse_filters(**filters))))),
                order_by=sql.SQL(", ".join([" ".join([i, order_by.generate()[i]]) for i in order_by.generate()])),
                limit=limit
            )
        cursor.execute(query=query)
        return cursor.fetchall()
    
    def delete(obj):
        cursor = DataBase().context.cursor()
        query: sql.SQL = sql.SQL("DELETE FROM {table} WHERE \"id\"={id};").format(
            table=sql.Identifier(obj.__class__.__base__().name),
            id=sql.Literal(obj.id)
        )
        cursor.execute(query=query)
        DataBase().context.commit()

    def save(obj):
        cursor = DataBase().context.cursor()
        fields = obj.__dict__
        set_: list = []
        for field in fields:
            if field == "id":
                continue
            else:
                set_.append(sql.SQL("{}={}").format(sql.Identifier(field), sql.Literal(fields[field])))
        query: sql.SQL = sql.SQL("UPDATE {} SET {} WHERE {};").format(
            sql.Identifier(obj.__class__.__base__().name), 
            sql.SQL(", ".join(list(map(lambda x: x.as_string(DataBase().context), set_)))),
            sql.SQL("id={}".format(obj.id))
        )
        cursor.execute(query=query)
        DataBase().context.commit()

    @classmethod
    def get(cls, id: int, schema: dataclass = None):
        cursor = None
        fields: list = []
        if schema:
            for field in schema.__annotations__:
                fields.append(sql.Identifier(field))
        else:
            for field in cls.fields():
                fields.append(sql.Identifier(field))
        if schema:
            cursor = DataBase().context.cursor(row_factory=class_row(schema))
        else:
            cursor = DataBase().context.cursor()
        query: sql.SQL = sql.SQL("SELECT {fields} FROM {table} WHERE id={id};").format(
            fields=sql.SQL(", ").join(fields),
            table=sql.Identifier(cls().name),
            id=id
        )
        cursor.execute(query=query)
        return cursor.fetchone()
    
    @classmethod
    def create(cls, **data):
        if data:
            cursor = DataBase().context.cursor()
            fields: list = []
            values: list = []
            for field in data:
                fields.append(sql.Identifier(field))
            for value in data:
                if isinstance(cls.fields().get(value), Password):
                    values.append(sql.Literal(hash_password(data[value])))
                else:
                    values.append(sql.Literal(data[value]))
            query: sql.SQL = sql.SQL("INSERT INTO {table}({fields}) VALUES ({values});").format(
                table=sql.Identifier(cls.__name__.lower() + "s"),
                fields=sql.SQL(", ").join(fields),
                values=sql.SQL(", ").join(values)
            )
            try:
                cursor.execute(query=query)
                DataBase().context.commit()
            except Exception as e:
                if isinstance(e, errors.NotNullViolation):
                    print("[ERROR]: {}".format(e))
                else:
                    print(e)
        else:
            print("[ERROR] no fields")

class Type:
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        self.null: bool = null
        self.unique: bool = unique
        self.primary_key: bool = primary_key
        self.is_array: bool = is_array
        self.default: str | int | float = default
        self.to_table: str = to_table
        self.to_column: str = to_column
        self.name: str = self.__class__.__name__.upper()

    def generate(self):
        params: list = []
        if self.is_array:
            self.name = self.name + "[]"
        params.append(self.name)
        if self.null:
            params.append("NULL")
        else:
            params.append("NOT NULL")
        if self.unique:
            params.append("UNIQUE")
        if self.primary_key:
            params.append("PRIMARY KEY")
        if self.default:
            re_compile = re.compile(r"\w+\(\)")
            re_match = re_compile.match(self.default)
            if re_match:
                params.append("DEFAULT {}".format(re_match.group(0)))
            elif type(self.default) == str:
                params.append("DEFAULT '{}'".format(self.default))
            elif type(self.default) == dict:
                params.append("DEFAULT '{}'".format(json.dumps(self.default)))
            else:
                params.append("DEFAULT {}".format(self.default))
        if self.to_table:
            if isinstance(self.to_table(), Model):
                if self.to_column:
                    params.append("REFERENCES {}({})".format(self.to_table().name, self.to_column))
                else:
                    params.append("REFERENCES {}(id)".format(self.to_table().name))
                    print("[WARNING]: to_column is not set, by default \"id\".")
            else:
                print("[ERROR]: to_table must be Model instance.")
        return " ".join(params)
    
class BigInt(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class BigSerial(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Boolean(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Date(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Integer(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Float(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class JSON(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Password(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)
        self.name = "VARCHAR(255)"
    
class Real(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class SmallSerial(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class SmallInt(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Serial(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Text(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class Time(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class TimeStamp(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class UUID(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = "gen_random_uuid()", to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

class VarChar(Type):
    def __init__(self, max_len: int, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        self.max_len: int = max_len
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)

    def generate(self):
        params: list = []
        if self.is_array:
            self.name = self.name + "[]"
        params.append("{}({})".format(self.name, self.max_len))
        if self.null:
            params.append("NULL")
        else:
            params.append("NOT NULL")
        if self.unique:
            params.append("UNIQUE")
        if self.primary_key:
            params.append("PRIMARY KEY")
        if self.to_table:
            if isinstance(self.to_table(), str):
                if self.to_column:
                    params.append("REFERENCES {}({})".format(self.to_table().name), self.to_column)
                else:
                    params.append("REFERENCES {}(id)".format(self.to_table().name))
                    print("[WARNING]: to_column is not set, by default \"id\".")
            else:
                print("[ERROR]: to_table must be Model instance.")
        return " ".join(params)
    
class XML(Type):
    def __init__(self, null: bool = True, unique: bool = False, primary_key: bool = False, is_array: bool = False, default: str | int | float = None, to_table: None = None, to_column: None = None):
        super().__init__(null, unique, primary_key, is_array, default, to_table, to_column)



import psycopg
from psycopg import sql

from kirpi.config import DNS

class DataBase:
    def __init__(self):
        self.context = None
        try:
            self.context = psycopg.connect(DNS)
        except:
            print("configurations is invalid, please check .config file")
            exit()


from psycopg import sql
from psycopg.rows import dict_row
import base64
import hashlib
import secrets
import re

from kirpi.base import DataBase


ALGORITHM = "pbkdf2_sha256"
EMAIL_REGEX = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'

def check_email(email):
    if (re.fullmatch(EMAIL_REGEX, email)):
        return True
    return False

def execute(query: str) -> tuple:
    cursor = DataBase().context.cursor(row_factory=dict_row)
    cursor.execute(query=query)
    DataBase().context.commit()
    return cursor.fetchall()

def hash_password(password: str, salt=None, iterations=260000):
    if salt is None:
        salt = secrets.token_hex(16)
    assert salt and isinstance(salt, str) and "$" not in salt
    assert isinstance(password, str)
    pw_hash = hashlib.pbkdf2_hmac(
        "sha256", password.encode("utf-8"), salt.encode("utf-8"), iterations
    )
    b64_hash = base64.b64encode(pw_hash).decode("ascii").strip()
    return "{}${}${}${}".format(ALGORITHM, iterations, salt, b64_hash)


def verify_password(password: str, password_hash: str):
    if (password_hash or "").count("$") != 3:
        return False
    algorithm, iterations, salt, b64_hash = password_hash.split("$", 3)
    iterations = int(iterations)
    assert algorithm == ALGORITHM
    compare_hash = hash_password(password, salt, iterations)
    return secrets.compare_digest(password_hash, compare_hash)

def hash_token(number: int):
    return secrets.token_hex(number)

def parse_filters(**kwargs):
    filters = []
    field = operator = ""
    for i in kwargs:
        if len(i.split("__")) == 2:
            field, operator = i.split("__")
            if operator == "ilike":
                filters.append(sql.SQL("{} ILIKE {}").format(sql.Identifier(field), sql.Literal(kwargs[i] + "%")))
            elif operator == "like":
                filters.append(sql.SQL("{} LIKE {}").format(sql.Identifier(field), sql.Literal(kwargs[i] + "%")))
            elif operator == "starts_with":
                filters.append(sql.SQL("{} ILIKE {}").format(sql.Identifier(field), sql.Literal(kwargs[i] + "%")))
            elif operator == "ends_with":
                filters.append(sql.SQL("{} ILIKE {}").format(sql.Identifier(field), sql.Literal("%" + kwargs[i])))
            elif operator == "lt":
                filters.append(sql.SQL("{} < {}").format(sql.Identifier(field), sql.Literal(kwargs[i])))
            elif operator == "gt":
                filters.append(sql.SQL("{} > {}").format(sql.Identifier(field), sql.Literal(kwargs[i])))
            elif operator == "lte":
                filters.append(sql.SQL("{} <= {}").format(sql.Identifier(field), sql.Literal(kwargs[i])))
            elif operator == "gte":
                filters.append(sql.SQL("{} >= {}").format(sql.Identifier(field), sql.Literal(kwargs[i])))
            elif operator == "ne":
                filters.append(sql.SQL("{} <> {}").format(sql.Identifier(field), sql.Literal(kwargs[i])))
        else:
            filters.append(sql.SQL("{} = {}").format(sql.Identifier(i), sql.Literal(kwargs[i])))
    return filters


from setuptools import setup, find_packages
setup(
    name='kirpi',
    version='1.0',
    author='Ali',
    author_email='alisher.abdimuminov.2005@gmail.com',
    description='Lightweight and simle ORM library for Python (with psycopg)',
    packages=find_packages(),
    ext_package=["psycopg"],
    license_file="LICENSE",
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)


import os
import click

database_config: str = """[DATABASE]
DB_HOST=<host>
DB_PORT=<port>
DB_NAME=<databasename>
DB_USER=<user>
DB_PASSWORD=<password>
"""

@click.command()
@click.argument("init")
def main(init) -> None:
    if init == "init":
        try:
            os.mkdir("models")
        except:
            pass
        try:
            os.mkdir("schemas")
        except:
            pass
        try:
            with open(".config", "w") as f:
                f.write(database_config)
        except:
            print("Do not create config file")
        print("Please edit .config file with your configurations.")
    else:
        raise click.BadArgumentUsage("{} is not defined".format(init))

if __name__ == "__main__":
    main()
